/***************************************************************************
*                                                                          *
*  Compact utility to install a Windows 3.1 application and set up a       *
*  progam object in Program Manager.                                       *                                                    *
*                                                                          *
*  August 1, 1996 - Public Domain by Kerry Imming                          *
*                                                                          *
*  Thanks to Mark Gamber for his "instex" package to create group/item     *
*  using DDE.  Public Domain by Mark Gamber, altered for use in this       *
*  package.                                                                *
*                                                                          *
****************************************************************************/

#include <windows.h>
#include <stdio.h>
#include <dos.h>
#include <dir.h>
#include <ctype.h>
#include <ddeml.h>
#include <string.h>
#include <stdlib.h>
#include <lzexpand.h>

#include "setup.h"

BOOL FAR PASCAL _export DlgProc(HWND, UINT, UINT, long) ;
BOOL GetProgmanGroups(HWND);

// ============================================================================

char INIFile[MAXPATHLEN];
char szInstallPath[MAXPATHLEN];
char szSourcePath[MAXPATHLEN];
char szCommandLine[MAXPATHLEN];
char szCaption[48];
char szDescription[48];
char szDefGroup[32];
char szMinSpace[16];
char szBuf[128];
HICON hIcon;


int PASCAL WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpszCmdLine, int iCmdShow)
  {
  //HWND        hwnd;
  //MSG         msg;
  FILE        *tstini;

  hIcon = LoadIcon(hInstance, IDI_SETUP);

  // Get current directory as default source directory
  getcwd(szSourcePath, MAXPATHLEN);

  // Create fully qualified path to INI file
  strcpy(INIFile, szSourcePath);
  // Append a trailing \ if needed
  if (INIFile[strlen(INIFile)-1] != '\\')
    strcat(INIFile, "\\");
  strcat(INIFile, "setup.ini");

  if ((tstini = fopen(INIFile, "rt")) == NULL)
    {
    strcpy(szBuf, "File not found: ");
    strcat(szBuf, INIFile);
    strcat(szBuf, " Install Failed.");
    MessageBox(NULL, szBuf, "Setup Error", MB_OK | MB_ICONEXCLAMATION);
    return 1;
    }
  else
    fclose(tstini);

  // Read parameters from setup.ini file
  GetPrivateProfileString("Setup", "Caption", "Setup",
     szCaption, sizeof(szCaption), INIFile);
  GetPrivateProfileString("Setup", "Description", "Unknown",
     szDescription, sizeof(szDescription), INIFile);
  GetPrivateProfileString("Setup", "CommandLine", "Unknown",
     szCommandLine, sizeof(szCommandLine), INIFile);
  GetPrivateProfileString("Setup", "DefGroup", "",
     szDefGroup, sizeof(szDefGroup), INIFile);
  GetPrivateProfileString("Setup", "DefDir", "C:\\TEMP",
     szInstallPath, sizeof(szInstallPath), INIFile);
  GetPrivateProfileString("Setup", "MinSpace", "0",
     szMinSpace, sizeof(szMinSpace), INIFile);

  if ( DialogBox( hInstance, MAKEINTRESOURCE( DLG_MAIN ), NULL, DlgProc) > 4)
     MessageBox(NULL,"Install Failed.", "Setup Error", MB_OK | MB_ICONEXCLAMATION);

  DestroyIcon(hIcon);

  return 0;
  }

// ============================================================================

long CopyFile(char szSrc[], char szDst[])
{
  long lrc;

  OFSTRUCT ofStrSrc;
  OFSTRUCT ofStrDest;
  HFILE hfSrcFile, hfDstFile;

  // Open the source file.
  hfSrcFile = LZOpenFile(szSrc, &ofStrSrc, OF_READ);

  // Create the destination file.
  hfDstFile = LZOpenFile(szDst, &ofStrDest, OF_CREATE);

  // Copy the source file to the destination file.
  lrc = LZCopy(hfSrcFile, hfDstFile);

  // Close the files.
  LZClose(hfSrcFile);
  LZClose(hfDstFile);

  return lrc;
}

// ============================================================================

int InstallFiles( HWND hDlg )
{
  FILE  *tstini;
  char  szDiskLbl[16];
  char  szDiskName[64];
  int   nDisk;
  PSTR  pszBuf, pszKey;
  char  szFile[80], szMsg[80];
  char  szSrcFile[MAXPATHLEN], szDstFile[MAXPATHLEN];
  HCURSOR hcurOrg, hcurWait;

  // Allocate a buffer for the entries.
  pszBuf = (PSTR) LocalAlloc(LMEM_FIXED, 4096);
  hcurWait = LoadCursor(NULL,IDC_WAIT);
  hcurOrg  = SetCursor(hcurWait);

  nDisk = 1;
  while ( TRUE ) {
    // Get Disk_n label, stop if not found
    sprintf(szDiskLbl,"Disk_%d",nDisk++);
    GetPrivateProfileString("Disks", szDiskLbl, "", szDiskName, sizeof(szDiskName), INIFile);
    if (!strcmp(szDiskName, "")) break;

    // Check for correct disk... prompt if not.
    strcpy(szSrcFile, szSourcePath);
    strcat(szSrcFile, szDiskLbl);
    while ( TRUE ) {
      if ((tstini = fopen(szSrcFile, "rt")) == NULL) {
        SetCursor(hcurOrg);
        strcpy(szMsg, "Insert disk: ");
        strcat(szMsg, szDiskName);
        if (MessageBox(NULL, szMsg, "Setup Message", MB_OKCANCEL) == IDCANCEL)
          {
          LocalFree((HANDLE) pszBuf);
          return 4;
          }
      }
      else {
        fclose(tstini);
        break;
      }
    }

    SetCursor(hcurWait);

    // Retrieve all the entries in the [Disk_n] section.
    GetPrivateProfileString(szDiskLbl, NULL, "", pszBuf, 4096, INIFile);

    // Retrieve the string for each entry, until
    // reaching the double null character.
      for (pszKey = pszBuf;
        *pszKey != '\0'; pszKey += strlen(pszKey) + 1) {

      // Retrieve the value for each entry in the buffer.
      GetPrivateProfileString(szDiskLbl, pszKey, "", szFile, sizeof(szFile), INIFile);

      while ( TRUE ) {
        strcpy(szSrcFile, szSourcePath);
        strcat(szSrcFile, szFile);

        strcpy(szDstFile, szInstallPath);
        strcat(szDstFile, szFile);

        strcpy(szMsg, "Copying: ");
        strcat(szMsg, szFile);
        SetDlgItemText( hDlg, IDD_STATUS, szMsg);

        if (CopyFile(szSrcFile, szDstFile) < 0)
          {
          SetCursor(hcurOrg);
          sprintf(szMsg, "Failure while copying \"%s\" to \"%s\"", szSrcFile, szDstFile);
          if (MessageBox(NULL, szMsg, "Setup Error",
              MB_RETRYCANCEL | MB_ICONEXCLAMATION) == IDCANCEL)
            {
            LocalFree((HANDLE) pszBuf);
            return 8;
            }
          SetCursor(hcurWait);
        }
        else
          break;
      }
    }
  }
  SetCursor(hcurOrg);
  strcpy(szMsg, " ");
  SetDlgItemText( hDlg, IDD_STATUS, szMsg);

  LocalFree((HANDLE) pszBuf);
  return 0;
}

/* Check disk space on the specified destination drive.
 *
 * Returns:
 *   TRUE if the destination provides sufficent disk space.
 *   FALSE if insufficient disk space.
 */
BOOL CheckDiskSpace( unsigned char ucDrive, long lNeeded )
{
  struct  diskfree_t free;
  long    lAvail;

  if (_dos_getdiskfree(ucDrive, &free) != 0)
    return FALSE;

  lAvail = (long) free.avail_clusters
         * (long) free.bytes_per_sector
         * (long) free.sectors_per_cluster;

  if (lAvail >= lNeeded)
    return TRUE;
  else
    return FALSE;
}

// ============================================================================

int CreateDir(PSTR szPath )
{
  char szDrive[_MAX_DRIVE];
  char szDir[_MAX_DIR];
  char szFile[_MAX_FNAME];
  char szExt[_MAX_EXT];
  char *pszDir;
  int  rc;
  int save, disk;

  // save original drive
  save = getdisk();

  // split the path to separate elements
  rc = fnsplit(szPath,szDrive,szDir,szFile,szExt);
  // Make sure drive and directory(s) ONLY specified
  if ( !(rc & DRIVE)     ||
       !(rc & DIRECTORY) ||
         rc & FILENAME   ||
         rc & EXTENSION )
    return 1;

  // Make sure the disk is valid
  setdisk(disk=toupper(szDrive[0])-'A');
  if (disk != getdisk())
    return 2;

  szDir[strlen(szDir)-1] = '\0';        // Delete Trailing \

  // If directory already exists... Done!
  chdir("\\");
  if ( !chdir(szDir) )
    return 0;

  // Get the first directory and cd to it
  chdir("\\");
  pszDir = strtok(szDir,"\\");
  if (pszDir)
    mkdir(pszDir);
  if ( (rc = chdir(pszDir)) != 0 )
    return rc;

  // Walk down the directory tree, creating? and cd'ing
  while ((pszDir = strtok(NULL,"\\")) != 0 ) {
    if (pszDir)
      mkdir(pszDir);
    if ( (rc = chdir(pszDir)) != 0 )
      return rc;
  }

  // Make sure it was really created
  rc = chdir(szDir);

  // Restore original drive
  setdisk(save);

  return rc;
}

// ============================================================================

// +--------------------------------------------------------------------------+
// | GetProgmanGroups() pulls a list of available groups from the Program     |
// |   Manager or something that supports similar DDE links. The list is cut  |
// |   into separate group names and added to the dialog's listbox.           |
// +--------------------------------------------------------------------------+
BOOL GetProgmanGroups( HWND hDlg )
{
   DWORD dwSize, dwDdeInst;
   HSZ Hsz;
   HCONV hConv;
   HDDEDATA hDdeData;
   HANDLE hMem;
   LPSTR lpStr;
   LPSTR lpNew;

   dwDdeInst = 0L;              //  This MUST be zeroed out before initializing
                                                   //  Initialize DDEML library
   if( DdeInitialize( &dwDdeInst, NULL, APPCLASS_STANDARD |
                      APPCMD_CLIENTONLY, 0 ) != DMLERR_NO_ERROR )
      return( FALSE );
                                          //  Same link and topic means one HSZ
   Hsz = DdeCreateStringHandle( dwDdeInst, "progman", 0 );
   hConv = DdeConnect( dwDdeInst, Hsz, Hsz, NULL );      //  Connect to PROGMAN
   DdeFreeStringHandle( dwDdeInst, Hsz );

   if( ! hConv )                                 //  If unable to connect, exit
   {
      DdeUninitialize( dwDdeInst );
      return( FALSE );
   }
                                                     //  Request current groups
   Hsz = DdeCreateStringHandle( dwDdeInst, "groups", 0 );
   hDdeData = DdeClientTransaction( NULL, 0, hConv, Hsz, CF_TEXT, XTYP_REQUEST,
                                    5000, NULL );
   DdeFreeStringHandle( dwDdeInst, Hsz );

   if( ! hDdeData )                               //  If nothing returned, exit
   {
      DdeDisconnect( hConv );
      DdeUninitialize( dwDdeInst );
      return( FALSE );
   }

   hMem = GlobalAlloc( GHND, 4096 );         //  Grab some space to hold string
   if( ! hMem )                                //  since it could be a long one
   {
      DdeDisconnect( hConv );               //  Groups are returned in a single
      DdeUninitialize( dwDdeInst );       //  string, each group name separated
      return( FALSE );                                //  by an '\x0D\x0A' pair
   }

   lpStr = GlobalLock( hMem );
   lpNew = DdeAccessData( hDdeData, &dwSize );           //  Lock down DDE data
   if( lpNew )
   {
      lstrcpy( lpStr, lpNew );           //  If lock worked, copy string to our
      DdeUnaccessData( hDdeData );             //  buffer and free DDEML's ASAP

      lpNew = _fstrchr( lpStr, '\x0D' );                 //  Look for separator
      while( *lpStr )
      {
         if( lpNew )                              //  If one found, NULL it out
            *lpNew = '\x00';                  //  to make a sendable group name
                                                //  Then, send the current name
         SendMessage( GetDlgItem( hDlg, IDD_GROUPLIST ), LB_ADDSTRING, 0,
                      (LPARAM)lpStr );
         if( lpNew )
         {                              //  If separator was found, replace and
            *lpNew = '\x0D';             //  point to two bytes beyond, that of
            lpStr = lpNew + 2;                          //  the next group name

            if( ! *lpStr )          //  If that's the end, break from this loop
               break;
                                                        //  Find next separator
            lpNew = _fstrchr( lpStr, '\x0D' );
         }
         else
            break;             //  If no separator found above, break from loop
      }
   }
   GlobalUnlock( hMem );             //  All done...back out the way we came in
   GlobalFree( hMem );        //  Note: hDdeData is killed off by DDEML, not us

   DdeDisconnect( hConv );                   //  Disconnect link and kill DDEML
   DdeUninitialize( dwDdeInst );

   return( TRUE );
}

// ============================================================================

// +--------------------------------------------------------------------------+
// | AddToProgman() adds the group or displays it if it already exists. It    |
// |      then adds this program to the shell, passing the filename, name to  |
// |      appear under the icon and icon number to use in the display.        |
// +--------------------------------------------------------------------------+
BOOL AddToProgman( HWND hDlg )
{
   char str[ 128 ];
   char cmnd[ 192 ];
   DWORD dwDdeInst;
   HCONV hConv;
   HSZ Hsz;
   int item;

   if( ! GetDlgItemText( hDlg, IDD_GROUP, str, sizeof(str)) )
   {
      MessageBox( hDlg, "You must enter a Group Name!", "Setup", MB_OK );
      return( FALSE );
   }

   dwDdeInst = 0L;
                                                   //  Initialize DDEML library
   if( DdeInitialize( &dwDdeInst, NULL, APPCLASS_STANDARD |
                      APPCMD_CLIENTONLY, 0 ) != DMLERR_NO_ERROR )
      return( FALSE );
                                          //  Same link and topic means one HSZ
   Hsz = DdeCreateStringHandle( dwDdeInst, "progman", 0 );
   hConv = DdeConnect( dwDdeInst, Hsz, Hsz, NULL );      //  Connect to PROGMAN
   DdeFreeStringHandle( dwDdeInst, Hsz );

   if( ! hConv )                                 //  If unable to connect, exit
   {
      DdeUninitialize( dwDdeInst );
      return( FALSE );
   }
                                              //  Is the group already created?
   item = SendMessage( GetDlgItem( hDlg, IDD_GROUPLIST ), LB_FINDSTRINGEXACT, -1,
                       (LPARAM)(LPSTR)str );
   if( item == LB_ERR )                       //  If not, send command to do so
   {
      sprintf(cmnd, "[CreateGroup(\"%s\")]\0", str);
   }
   else                    //  Otherwise, display existing group if not already
   {
      SendMessage( GetDlgItem( hDlg, IDD_GROUPLIST ), LB_GETTEXT, item,
                   (LPARAM)(LPSTR)str );
      sprintf(cmnd, "[ShowGroup(\"%s\",1)]", str);
   }
                                       //  Either way, send the DDE transaction
   if( ! DdeClientTransaction( cmnd, lstrlen( cmnd ) + 1, hConv, NULL,
                               CF_TEXT, XTYP_EXECUTE, 5000, NULL ) )
   {
      DdeDisconnect( hConv );        //  If something failed, clean up and exit
      DdeUninitialize( dwDdeInst );
      return( FALSE );
   }
                                         //  Set up command to add this program
   sprintf(cmnd, "[AddItem(%s%s,\"%s\",,%d)]\0", szInstallPath, szCommandLine,
       szDescription, 0);

                                                       //  Blast it through DDE
   if( ! DdeClientTransaction( cmnd, lstrlen( cmnd ) + 1, hConv, NULL,
                               CF_TEXT, XTYP_EXECUTE, 5000, NULL ) )
   {
      DdeDisconnect( hConv );
      DdeUninitialize( dwDdeInst );        //  Reflect failure on return
      return( FALSE );
   }

   DdeDisconnect( hConv );
   DdeUninitialize( dwDdeInst );           //  Reflect success on return
   return( TRUE );
}

// ============================================================================

int DoInstall (HWND hwnd)
{
  int rc;

  // Append a trailing \ unless the user already did
  if (szInstallPath[strlen(szInstallPath)-1] != '\\')
    strcat(szInstallPath, "\\");

  // Append a trailing \ unless the user already did
  if (szSourcePath[strlen(szSourcePath)-1]  != '\\')
    strcat(szSourcePath, "\\");

  if (szInstallPath[1] != ':')
    {
    MessageBox(NULL, "Invalid drive specified for install target.",
               "Setup Error", MB_OK | MB_ICONEXCLAMATION );
    return 8;
    }

  if (!CheckDiskSpace(toupper(szInstallPath[0])-'A'+1, atol(szMinSpace)) )
    {
    MessageBox(NULL, "Not enough space on target drive or target drive is invalid.",
               "Setup Error", MB_OK | MB_ICONEXCLAMATION );
    return 8;
    }

  if ((rc = CreateDir(szInstallPath)) != 0 )
    {
    sprintf(szBuf, "Error creating target directory: %s  ", szInstallPath);
    if (rc == -1)
      strcat(szBuf, strerror(errno));
    MessageBox(NULL, szBuf, "Setup Error", MB_OK | MB_ICONEXCLAMATION);
    return 8;
    }

  if ((rc = InstallFiles(hwnd)) != 0 )
    return rc;

  AddToProgman( hwnd );                 // Add item to shell
  return 0;
}

// ============================================================================

BOOL FAR PASCAL _export DlgProc (HWND hwnd, UINT iMsg, UINT wParam, long lParam)
     {
  int  index;
  HDC  hdc;
  PAINTSTRUCT ps;

  switch( iMsg )
  {
    case WM_INITDIALOG:
      SetWindowText ( hwnd, szCaption);
      SetDlgItemText( hwnd, IDD_SRCDIR, szSourcePath);
      SetDlgItemText( hwnd, IDD_DSTDIR, szInstallPath);
      SetDlgItemText( hwnd, IDD_GROUP,  szDefGroup);

      if(!GetProgmanGroups( hwnd ))
        {
        // Unable to locate groups from shell, exit!
        MessageBox(NULL, "Failure getting groups from Program Manager.",
                       "Setup Error", MB_OK | MB_ICONEXCLAMATION);
        EndDialog( hwnd, 8 );
        }
      return TRUE;

    case WM_PAINT:
      if (IsIconic(hwnd))
        {
        hdc = BeginPaint(hwnd, &ps);
        DrawIcon(hdc, 0, 0, hIcon);
        EndPaint(hwnd, &ps);
        return TRUE;
        }

    case WM_COMMAND:
      if( wParam == IDCANCEL )
      {
        EndDialog( hwnd, 4 );
        return TRUE;
      }

      if( wParam == IDOK )
        {
        GetDlgItemText( hwnd, IDD_SRCDIR, szSourcePath, sizeof(szSourcePath));
        GetDlgItemText( hwnd, IDD_DSTDIR, szInstallPath, sizeof(szInstallPath));

        EndDialog( hwnd, DoInstall(hwnd) );
        return TRUE;
        }

      if( HIWORD( lParam ) == LBN_SELCHANGE )     //  Listbox item selection
      {
        //  Get the selection text and put it in the edit control
        index = SendMessage( GetDlgItem( hwnd, IDD_GROUPLIST ), LB_GETCURSEL, 0, 0 );
        if( index == LB_ERR )
          break;

        if( ! SendMessage( GetDlgItem( hwnd, IDD_GROUPLIST ), LB_GETTEXT, index,
                           (LPARAM)(LPSTR)szBuf ) )
           break;
        SetDlgItemText( hwnd, IDD_GROUP, szBuf );
        break;
      }
      break;

   }

   return FALSE;
 }
